import pytest
import inspect
from struphy.tests.test_codes.util import call_model


@pytest.mark.mpi(min_size=2)
@pytest.mark.parametrize('map_and_equil', [('Cuboid', 'HomogenSlab'),
                                           ('HollowTorus', 'AdhocTorus'),
                                           ('Tokamak', 'EQDSKequilibrium')
                                           ])
def test_toy(fast, map_and_equil, model=None):
    '''Tests all models and all possible model.options (except solvers without preconditioner) in models/toy.py.

    If model is not None, tests the specified model.

    The argument "fast" is a pytest option that can be specified at te command line (see conftest.py).'''

    from struphy.models import toy

    if model is None:
        for key, val in inspect.getmembers(toy):
            if inspect.isclass(val) and 'StruphyModel' not in key:

                if fast:
                    if 'Cuboid' not in map_and_equil[0]:
                        print(
                            f'Fast is enabled, mapping {map_and_equil[0]} skipped ...')
                        continue

                call_model(key, val, map_and_equil)
    else:
        val = getattr(toy, model)
        call_model(model, val, map_and_equil)
